output_dir = "~/tmp/tani-cost/";

log2(x) = {if(x == 0, 0, log(x)/log(2))};

isogeny_cost(log2_p) = {
  my(log2_G, log2_W);
  e = log2_p/2;
  Fp2mul_G = 4*log2_p*log2(log2_p); \\ 1 F_{p^2} mul is 4 F_p mul
  log2_G = log2(e/2 * log2(e/2) * Fp2mul_G);
  log2_W = log2(2*log2_p);
  [log2_G, log2_G - log2_W, log2_W];
};

isogeny_claw_storage(log2_p) = {
  \\ Store (path, Hash(f(path)))
  \\ Path is (log p)/4 bits. Hash should be at least the same size.
  log2(1/2 * log2_p);
}

cost_johnson_vertex_claw_count(log2_storage, log2_R) = {
  my(log2_gates, log2_depth, log2_width);
  \\ storage is the size of an element, R is the set size

  \\ Cost 3:
  \\ gates = R m 
  \\ depth = log m + log^2 R
  \\ width = R m

  log2_gates = log2_R + log2_storage;
  log2_depth = vecmax([log2(log2_storage), 2*log2(log2_R)]);
  log2_width = log2_R + log2_storage;

  [log2_gates, log2_depth, log2_width];
}

cost_mnrs_random_walk(log2_S_costs, log2_C_costs, log2_U_costs, log2_eps, log2_delta) = {
  my(log2_SG, log2_SD, log2_SW, log2_CG, log2_CD, log2_CW, log2_UG, log2_UD, log2_UW);
  my(log2_G, log2_D, log2_W);

  \\ Cost 5:
  \\ gates = SG + 1/sqrt(eps)(1/sqrt(delta) UG + CG)
  \\ depth = SD + 1/sqrt(eps)(1/sqrt(delta) UD + CD)
  \\ width = max(SW, UW - log(eps*delta), CW - log(eps*delta))

  [log2_SG, log2_SD, log2_SW] = log2_S_costs;
  [log2_CG, log2_CD, log2_CW] = log2_C_costs;
  [log2_UG, log2_UD, log2_UW] = log2_U_costs;

  log2_G = vecmax([ log2_SG, \
                    log2_UG - (log2_eps + log2_delta)/2, \
                    log2_CG - log2_eps/2]);

  log2_D = vecmax([ log2_SD, \
                    log2_UD - (log2_eps + log2_delta)/2, \
                    log2_CD - log2_eps/2]);

  log2_W = vecmax([ log2_SW, \
                    max(log2_UW, log2(-(log2_eps+log2_delta))), \
                    max(log2_CW, log2(-(log2_eps+log2_delta)))]);

  [log2_G, log2_D, log2_W];
}

mnrs_log2_eps(log2_X, log2_R) = {
  \\ epsilon = R^2 / |X|^2
  log2_eps = 2*(log2_R - log2_X);
}

mnrs_log2_delta(log2_X, log2_R) = {
  \\ delta = 1/R
  log2_delta = -log2_R;
}

cost_mnrs_golden_claw(log2_X, log2_R, log2_sample_costs, log2_storage) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W);
  my(log2_insert_G, log2_insert_D, log2_insert_W);
  my(log2_UG,log2_UD,log2_UW, log2_SG, log2_SD, log2_SW);
  my(log2_S_costs, log2_C_costs, log2_eps, log2_delta);

  \\ The set-up and update costs can be expressed in terms of
  \\ the cost of sampling from X and the cost of inserting into
  \\ a list of size R.
  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;
  [log2_insert_G, log2_insert_D, log2_insert_W] = cost_johnson_vertex_claw_count(log2_storage, log2_R);

  \\ Update cost is ~ 2 samples and ~ 2 insertions. We only charge for 1 of each.
  log2_UG = max(log2_sample_G, log2_insert_G);
  log2_UD = max(log2_sample_D, log2_insert_D);
  log2_UW = max(log2_sample_W, log2_insert_W);
  log2_U_costs = [log2_UG, log2_UD, log2_UW];

  \\ Setup cost ~ R updates
  log2_SG = log2_R + log2_UG;
  log2_SD = log2_R + log2_UD;

  \\ Width is dominated by total storage (2 R m) or by
  \\ one of the subroutine
  log2_SW = vecmax([1 + log2_R + log2_storage, \
                    log2_sample_W, \
                    log2_insert_W]);
  log2_S_costs = [log2_SG, log2_SD, log2_SW];

  \\ Check cost is O(1) depth and gates, R m width
  log2_C_costs = [0, 0, log2_R + log2_storage];

  log2_eps = mnrs_log2_eps(log2_X,log2_R);
  log2_delta = mnrs_log2_delta(log2_X, log2_R);

  cost_mnrs_random_walk(log2_S_costs, log2_C_costs, log2_U_costs, log2_eps, log2_delta);
}


\\ Tani's algorithm finds a claw between functions f and g.
\\ using an MNRS random walk on a product of Johnson graphs
\\    J(dom(f),R_f) x J(dom(g),R_g)
\\ Here we assume |dom(f)| = |dom(g)| = X and R_f = R_g = R.
\\ For an attack on SIKE X := p^{1/4}.
\\ Basic parallelization strategy is to divide each of the domains
\\ into S parts, S^2 parts total, then distribute pairs of parts among
\\ S^2 processors. Each processor uses Tani's algorithm and stores R elements
\\ of each domain in memory (2R elements total).

data_tani_with_maxdepth(log2_p, log2_sample_costs, maxdepth) = {
  my(log2_X, log2_R, log2_S, log2_G, log2_D, log2_W);

  \\ We impose a depth bound of 2^maxdepth gates and minimize memory usage.

  log2_storage = isogeny_claw_storage(log2_p);

  \\ Start with the gate-optimal choice of R.
  log2_X = (1/4)*log2_p;
  log2_R = log2_sample_costs[1] - log2_storage;
  log2_S = 0;

  [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage);

  if(log2_D < maxdepth,
    \\ If depth is less than maxdepth, decrease R.
    while(ceil(log2_D) < maxdepth && log2_R >= 1,
      log2_R -= 1;
      [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage)),
    \\ Otherwise parallelize.
    while(ceil(log2_D) > maxdepth && log2_X > log2_S,
      log2_S = min(log2_S+1, log2_X);
      if(log2_R > (2/3)*(log2_X - log2_S), log2_R = (2/3)*(log2_X - log2_S));
      [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage)));

  \\ Account for S^2 parallelism in final cost
  [log2_p, log2_G+2*log2_S, log2_D, log2_W+2*log2_S, 2*log2_S, log2_R];
}

data_tani_with_maxmemory(log2_p, log2_sample_costs, maxmemory) = {
  my(log2_X, log2_R, log2_S, log2_G, log2_D, log2_W);

  \\ We impose a memory bound of 2^maxmemory bits and minimize depth.

  log2_storage = isogeny_claw_storage(log2_p);

  \\ Start with the gate-optimal choice of R.
  log2_X = (1/4)*log2_p;
  log2_R = log2_sample_costs[1] - log2_storage;
  log2_S = 0;

  [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage);

  if(ceil(log2_W) < maxmemory,
    \\ If we have extra memory, parallelize
    while((ceil(log2_W+2*log2_S) < maxmemory) && (log2_X > log2_S),
      log2_S = min(log2_S+1, log2_X);
      if(log2_R > (2/3)*(log2_X - log2_S), log2_R = (2/3)*(log2_X - log2_S));
      [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage)),
    \\ Otherwise decrease R until we're below maxmemory.
    while((ceil(log2_W+2*log2_S) > maxmemory) && (log2_R >= 1),
      log2_R -= 1;
      [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage)));

  [log2_p, log2_G+2*log2_S, log2_D, log2_W+2*log2_S, 2*log2_S, log2_R];
}

data_tani_min_gates(log2_p,log2_sample_costs) = {
  my(log2_X, log2_R, log2_S, log2_G, log2_D, log2_W,prev_G);

  \\ We start with R=0 and increase R until it hits a minimum G cost
  \\ Assumes a single global minimum at a positive R
  log2_X = (1/4)*log2_p;
  log2_R = 0;
  log2_S = 0;

  log2_storage = isogeny_claw_storage(log2_p);

  [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage);
  prev_G=log2_G+1;
  while(prev_G>log2_G,
    prev_G=log2_G;
    log2_R += 1;
    [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage));
  log2_R -= 1;
  [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage);
  [log2_p, log2_G+2*log2_S, log2_D, log2_W+2*log2_S, 2*log2_S, log2_R];
}

data_tani_min_dw(log2_p,log2_sample_costs) = {
  my(log2_X, log2_R, log2_S, log2_G, log2_D, log2_W,prev_G);

  \\ We start with R=0 and increase R until it hits a minimum DW cost
  \\ Assumes a single global minimum at a positive R
  log2_X = (1/4)*log2_p;
  log2_R = 0;
  log2_S = 0;

  log2_storage = isogeny_claw_storage(log2_p);

  [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage);
  prev_DW=log2_D+log2_W+1;
  while(prev_DW>log2_D+log2_W,
    prev_DW=log2_D + log2_W;
    log2_R += 1;
    [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage));
  log2_R -= 1;
  [log2_G, log2_D, log2_W] = cost_mnrs_golden_claw(log2_X - log2_S, log2_R, log2_sample_costs, log2_storage);
  [log2_p, log2_G+2*log2_S, log2_D, log2_W+2*log2_S, 2*log2_S, log2_R];
}


cost_grover(log2_p, log2_sample_costs, log2_processors) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W, log2_X, log2_G, log2_D);

  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;

  log2_X = (1/4)*log2_p;
  log2_G = (1/2)*(2*log2_X + log2_processors) + log2_sample_G;
  log2_D = (1/2)*(2*log2_X - log2_processors) + log2_sample_D;
  [log2_G, log2_D];
}

data_grover_with_maxdepth(log2_p, log2_sample_costs, maxdepth) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W, log2_W, log2_S, log2_G, log2_D);

  \\ We impose a depth bound of 2^maxdepth gates and minimize memory.

  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;

  \\ Sequential grover is gate and memory optimal, but has large depth.
  log2_S = 0;
  [log2_G, log2_D] = cost_grover(log2_p, log2_sample_costs, log2_S);

  \\ Increase processors until we're below maxdepth.
  while(ceil(log2_D) > maxdepth,
    log2_S += 1;
    [log2_G, log2_D] = cost_grover(log2_p, log2_sample_costs, log2_S));

  log2_W = log2_S + log2_sample_W;

  [log2_p, log2_G, log2_D, log2_W, log2_S];
}

data_grover_with_maxmemory(log2_p, log2_sample_costs, maxmemory) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W, log2_W, log2_S, log2_G, log2_D);

  \\ We impose a memory bound of 2^maxmemory bits and minimize depth.

  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;

  \\ Assume we can use all the memory.
  log2_S = maxmemory - log2_sample_W;
  [log2_G, log2_D] = cost_grover(log2_p, log2_sample_costs, log2_S);

  \\ Back off until each processor makes at least one query.
  while(log2_D < log2_sample_D && log2_S > 1,
    log2_S -= 1;
    if (log2_S<1,log2_S=0);
    [log2_G, log2_D] = cost_grover(log2_p, log2_sample_costs, log2_S));

  \\ Memory usage might have changed.
  log2_W = log2_S + log2_sample_W;

  [log2_p, log2_G, log2_D, log2_W, log2_S];
}



cost_vw(log2_p, log2_sample_costs, log2_storage, log2_W, log2_S) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W, log2_G, log2_D, log2_processors, log2_memory_units);

  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;

  \\ Cannot use more than m * p^{1/4} bits of memory.
  log2_W = min(log2_W, 1/4 * log2_p + log2_storage);

  \\ VW cost formula assumes each unit of memory can store a distinguished point
  log2_memory_units = log2_W - log2_storage;
  log2_G = (3/8)*log2_p - (1/2)*log2_memory_units + log2_sample_G;

  \\ Perfect parallelism (this function can return a nonsensical depth, must be checked).
  log2_D = log2_G - log2_S;

  \\ If we have a huge number of processors they might dominate memory usage.
  log2_W = max(log2_W, log2_S + log2_sample_W);

  [log2_G, log2_D, log2_W];
}

data_vw_with_maxdepth(log2_p,log2_sample_costs,maxdepth) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W, log2_G, log2_D, log2_W, log2_S);

  \\ We impose a depth bound of 2^maxdepth gates and minimize memory.

  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;

  \\ Start with the minimum amount of memory possible.
  \\ Assume processors proportional to memory.
  log2_W = log2_sample_W;
  log2_storage = isogeny_claw_storage(log2_p);
  log2_S = log2_W - max(log2_sample_W, log2_storage);
  [log2_G, log2_D, log2_W] = cost_vw(log2_p, log2_sample_costs, log2_storage, log2_W, log2_S);

  \\ Increase memory until we're below maxdepth
  while(ceil(log2_D) > maxdepth,
    log2_W += 1;
    log2_S = log2_W - max(log2_sample_W, log2_storage);
    [log2_G, log2_D, new_log2_W] = cost_vw(log2_p, log2_sample_costs, log2_storage, log2_W, log2_S);
    if(new_log2_W < log2_W, break));

  \\ Max gates with p^{1/4 + o(1)}, here I'm assuming the o(1) masks log2_sample_gates
  log2_G = max(log2_G, (1/4)*log2_p + log2_sample_G); 
  [log2_p, log2_G, log2_D, log2_W, log2_S];
}

data_vw_with_maxmemory(log2_p,log2_sample_costs,maxmemory) = {
  my(log2_sample_G, log2_sample_D, log2_sample_W, log2_G, log2_D, log2_W, log2_S);

  \\ We impose a memory bound of 2^maxmemory bits and minimize depth.

  [log2_sample_G, log2_sample_D, log2_sample_W] = log2_sample_costs;

  \\ Assume we can use all of the memory.
  log2_W=maxmemory;
  log2_storage = isogeny_claw_storage(log2_p);
  log2_S = log2_W - max(log2_sample_W, log2_storage);
  [log2_G, log2_D, log2_W] = cost_vw(log2_p, log2_sample_costs, log2_storage, log2_W, log2_S);

  \\ Back off until each processor makes at least one query.
  while(log2_D < log2_sample_D && log2_S > 1,
    log2_S -= 1;
    if (log2_S<1,log2_S=0);
    [log2_G, log2_D, log2_W] = cost_vw(log2_p, log2_sample_costs, log2_storage, log2_W, log2_S));

  \\ Max gates with p^{1/4 + o(1)}, here I'm assuming the o(1) masks log2_sample_gates
  log2_G = max(log2_G, (1/4)*log2_p + log2_sample_G); 
  [log2_p, log2_G, log2_D, log2_W, log2_S];
}
